import torchmix.components as components
import torchmix.nn as nn
from torchmix.core._context import no_parameters
from torchmix.core._module import MixModule
from torchmix.components import *

# __all__ = [
#     "MixModule",
#     "no_parameters",
#     "nn",
# ]



import functools
import inspect
import os
import re
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Optional

from hydra.core.config_store import ConfigStore
from hydra_zen import instantiate
from hydra_zen.typing import Partial
from omegaconf import DictConfig, OmegaConf
from torch import nn
from typing_extensions import Self

from torchmix.core._builds import BuildMode, builds

NO_PARAMS = False


def _underscore(name: str) -> str:
    name = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", name)
    name = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", name)
    name = name.replace("-", "_")
    return name.lower()


def _parse(value):
    if isinstance(value, MixModule):
        return value.config
    elif isinstance(value, Partial):
        return builds(
            value.func,
            *value.args,
            **value.keywords,
            zen_partial=True,
        )
    elif isinstance(value, Mapping):
        return builds(
            type(value),
            {k: _parse(v) for k, v in value.items()},
            populate_full_signature=False,
        )
    elif isinstance(value, str):
        return value
    elif isinstance(value, Sequence):
        return builds(
            type(value),
            [_parse(elem) for elem in value],
        )

    return value


class MixModule(nn.Module):
    """
    A PyTorch module wrapper that automates the process of generating and
    managing configurations for PyTorch modules.

    This class inherits from PyTorch's `nn.Module` and can be used as
    a drop-in replacement. It automatically generates configurations
    for any instance of the `MixModule` class or its subclasses and
    stores them in a `DictConfig` object.

    The generated configurations can be accessed via the `config` property and
    exported to a file using the `export` method.

    The `store` method can be used to store the configurations in
    `hydra`'s ConfigStore for use with `hydra`'s CLI.
    """

    _config: DictConfig
    _option_name: str

    build_mode: BuildMode = BuildMode.WITHOUT_ARGS

    def __new__(cls, *args, **kwargs):
        """
        Overrides the default behavior of `MixModule`'s constructor. This
        method is called before the actual `__init__` method and is used to
        build the `config` attribute of the `MixModule` instance.

        This method is responsible for generating the configuration for a
        `MixModule` instance before calling the original `__init__` method.
        It does this by creating a new `__init__` method that calls
        `self.builds` with the provided arguments after recursively parsing
        them with the `_parse` function.

        The `_parse` function converts any `MixModule` instances found in
        the arguments to their corresponding configurations, as well as
        handling any partial instances and wrapping dictionaries and sequences
        in their respective types.

        The resulting configuration is then stored in `self._config`. The new
        `__init__` method then calls the original `__init__` method with the
        provided arguments.

        When `no_parameters` context manager is active, this method will
        skip calling the original `__init__` method, and the `MixModule`
        instance will not have any additional attributes or behavior beyond
        the `config` attribute.

        This can be useful if you want to create a `MixModule` instance
        solely for the purpose of generating its configuration, without
        actually instantiating the module. For more details, see
        `torchmix.core._context.no_parameters` context manager.

        Returns:
            The `MixModule` instance.
        """
        old_init = cls.__init__
        old_sig = inspect.signature(old_init)

        def new_init(self, *args, **kwargs):
            _args = ()
            _kwargs = {}
            for arg in args:
                _args += (_parse(arg),)

            for key, value in kwargs.items():
                _kwargs[key] = _parse(value)

            self._config = self.builds(*_args, **_kwargs)

            if not NO_PARAMS:
                nn.Module.__init__(self)
                old_init(self, *args, **kwargs)

            cls.__init__ = old_init

        new_init.__signature__ = old_sig.replace(
            parameters=list(old_sig.parameters.values())
        )

        cls.__init__ = new_init
        return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def builds(cls, *args, **kwargs):
        """
        Generates a configuration for an instance of the `MixModule` class
        or its subclasses.

        This method is used internally to generate the configuration for
        a `MixModule` instance. It should not be called directly.

        Returns:
            A `DictConfig` object containing the generated configuration for
            the instance.
        """
        if cls.build_mode is BuildMode.WITHOUT_ARGS:
            return builds(cls, **kwargs)(*args)
        elif cls.build_mode is BuildMode.WITH_ARGS:
            return builds(cls, *args, **kwargs)
        else:
            raise ValueError("build_mode must be one of WITHOUT_ARGS or WITH_ARGS")

    @classmethod
    def partial(cls, *args, **kwargs):
        """
        Returns a `functools.partial` object that can be used to partially
        instantiate the `MixModule` class or its subclasses.

        Returns:
            A `functools.partial` object that can be called like a function to
            create a `MixModule` instance.
        """
        return functools.partial(cls, *args, **kwargs)

    @property
    def config(self):
        """
        Returns the `DictConfig` object containing the configuration for the
        `MixModule` instance.

        This property allows users to access the generated configuration for
        the `MixModule` instance.
        """
        return self._config

    def export(
        self,
        path: Optional[str] = None,
    ) -> Self:  # type: ignore
        """
        Exports the configuration for the `MixModule` instance to a file.

        This method allows users to save the configuration for the
        `MixModule` instance to a file in YAML format.

        If no path is provided, the configuration will be saved to
        a file with the name of the class in the current working directory.

        Parameters:
            path: Path to the file where the configuration should be saved.

        Returns:
            The `MixModule` instance.
        """

        if path:
            _, extension = os.path.splitext(path)
            if extension:
                path = path
            else:
                if not self.option_name:
                    path = f"{path}.yaml"
                else:
                    Path(path).mkdir(parents=True, exist_ok=True)
                    path = f"{path}/{self.option_name}.yaml"
        else:
            if not self.option_name:
                path = path = f"{_underscore(type(self).__name__)}.yaml"
            else:
                path = self.option_name + ".yaml"

        OmegaConf.save(self.config, path)

        return self

    @property
    def option_name(self) -> Optional[str]:
        try:
            return self._option_name
        except AttributeError:
            return None

    def store(self, group: str, name: str) -> Self:  # type: ignore
        """
        Stores the configuration for the `MixModule` instance in
        `hydra`'s ConfigStore.
        """
        self._option_name = name

        cs = ConfigStore.instance()
        cs.store(group=group, name=name, node=self.config)

        return self

    def instantiate(self) -> Self:  # type: ignore
        """
        Recreates an instance of the MixModule from its configuration.

        This method recreates an instance of the `MixModule` by calling
        `hydra_zen.instantiate` on the `config` attribute of the
        `MixModule` instance. It returns the recreated instance.

        This can be useful if you want to instantiate a `MixModule`
        multiple times using the same configuration.

        Returns:
            The recreated instance of the `MixModule`.
        """
        return instantiate(self.config)



"""
This file is automatically generated by torchmix and should not be changed.

This module contains wrapped versions of the objects in the PyTorch `nn`
module, intended to provide a more user-friendly interface and added
configuration management functionality.

These classes can be used as a drop-in replacement for the original PyTorch
`nn` module, allowing you to access the full configuration of the module
through the config attribute.
"""

import functools
from contextlib import suppress

from torch import nn

from torchmix.core._builds import BuildMode
from torchmix.core._module import MixModule

Module = nn.Module
Parameter = nn.Parameter
UninitializedBuffer = nn.UninitializedBuffer
UninitializedParameter = nn.UninitializedParameter

with suppress(AttributeError):
    factory_kwargs = nn.factory_kwargs

with suppress(AttributeError):
    functional = nn.functional

with suppress(AttributeError):
    init = nn.init

with suppress(AttributeError):
    modules = nn.modules

with suppress(AttributeError):
    parallel = nn.parallel

with suppress(AttributeError):
    parameter = nn.parameter

with suppress(AttributeError):
    utils = nn.utils


with suppress(AttributeError):

    class AdaptiveAvgPool1d(MixModule, nn.AdaptiveAvgPool1d):
        __init__ = functools.partial(nn.AdaptiveAvgPool1d.__init__)


with suppress(AttributeError):

    class AdaptiveAvgPool2d(MixModule, nn.AdaptiveAvgPool2d):
        __init__ = functools.partial(nn.AdaptiveAvgPool2d.__init__)


with suppress(AttributeError):

    class AdaptiveAvgPool3d(MixModule, nn.AdaptiveAvgPool3d):
        __init__ = functools.partial(nn.AdaptiveAvgPool3d.__init__)


with suppress(AttributeError):

    class AdaptiveLogSoftmaxWithLoss(MixModule, nn.AdaptiveLogSoftmaxWithLoss):
        __init__ = functools.partial(nn.AdaptiveLogSoftmaxWithLoss.__init__)


with suppress(AttributeError):

    class AdaptiveMaxPool1d(MixModule, nn.AdaptiveMaxPool1d):
        __init__ = functools.partial(nn.AdaptiveMaxPool1d.__init__)


with suppress(AttributeError):

    class AdaptiveMaxPool2d(MixModule, nn.AdaptiveMaxPool2d):
        __init__ = functools.partial(nn.AdaptiveMaxPool2d.__init__)


with suppress(AttributeError):

    class AdaptiveMaxPool3d(MixModule, nn.AdaptiveMaxPool3d):
        __init__ = functools.partial(nn.AdaptiveMaxPool3d.__init__)


with suppress(AttributeError):

    class AlphaDropout(MixModule, nn.AlphaDropout):
        __init__ = functools.partial(nn.AlphaDropout.__init__)


with suppress(AttributeError):

    class AvgPool1d(MixModule, nn.AvgPool1d):
        __init__ = functools.partial(nn.AvgPool1d.__init__)


with suppress(AttributeError):

    class AvgPool2d(MixModule, nn.AvgPool2d):
        __init__ = functools.partial(nn.AvgPool2d.__init__)


with suppress(AttributeError):

    class AvgPool3d(MixModule, nn.AvgPool3d):
        __init__ = functools.partial(nn.AvgPool3d.__init__)


with suppress(AttributeError):

    class BCELoss(MixModule, nn.BCELoss):
        __init__ = functools.partial(nn.BCELoss.__init__)


with suppress(AttributeError):

    class BCEWithLogitsLoss(MixModule, nn.BCEWithLogitsLoss):
        __init__ = functools.partial(nn.BCEWithLogitsLoss.__init__)


with suppress(AttributeError):

    class BatchNorm1d(MixModule, nn.BatchNorm1d):
        __init__ = functools.partial(nn.BatchNorm1d.__init__)


with suppress(AttributeError):

    class BatchNorm2d(MixModule, nn.BatchNorm2d):
        __init__ = functools.partial(nn.BatchNorm2d.__init__)


with suppress(AttributeError):

    class BatchNorm3d(MixModule, nn.BatchNorm3d):
        __init__ = functools.partial(nn.BatchNorm3d.__init__)


with suppress(AttributeError):

    class Bilinear(MixModule, nn.Bilinear):
        __init__ = functools.partial(nn.Bilinear.__init__)


with suppress(AttributeError):

    class CELU(MixModule, nn.CELU):
        __init__ = functools.partial(nn.CELU.__init__)


with suppress(AttributeError):

    class CTCLoss(MixModule, nn.CTCLoss):
        __init__ = functools.partial(nn.CTCLoss.__init__)


with suppress(AttributeError):

    class ChannelShuffle(MixModule, nn.ChannelShuffle):
        __init__ = functools.partial(nn.ChannelShuffle.__init__)


with suppress(AttributeError):

    class ConstantPad1d(MixModule, nn.ConstantPad1d):
        __init__ = functools.partial(nn.ConstantPad1d.__init__)


with suppress(AttributeError):

    class ConstantPad2d(MixModule, nn.ConstantPad2d):
        __init__ = functools.partial(nn.ConstantPad2d.__init__)


with suppress(AttributeError):

    class ConstantPad3d(MixModule, nn.ConstantPad3d):
        __init__ = functools.partial(nn.ConstantPad3d.__init__)


with suppress(AttributeError):

    class Container(MixModule, nn.Container):
        __init__ = functools.partial(nn.Container.__init__)


with suppress(AttributeError):

    class Conv1d(MixModule, nn.Conv1d):
        __init__ = functools.partial(nn.Conv1d.__init__)


with suppress(AttributeError):

    class Conv2d(MixModule, nn.Conv2d):
        __init__ = functools.partial(nn.Conv2d.__init__)


with suppress(AttributeError):

    class Conv3d(MixModule, nn.Conv3d):
        __init__ = functools.partial(nn.Conv3d.__init__)


with suppress(AttributeError):

    class ConvTranspose1d(MixModule, nn.ConvTranspose1d):
        __init__ = functools.partial(nn.ConvTranspose1d.__init__)


with suppress(AttributeError):

    class ConvTranspose2d(MixModule, nn.ConvTranspose2d):
        __init__ = functools.partial(nn.ConvTranspose2d.__init__)


with suppress(AttributeError):

    class ConvTranspose3d(MixModule, nn.ConvTranspose3d):
        __init__ = functools.partial(nn.ConvTranspose3d.__init__)


with suppress(AttributeError):

    class CosineEmbeddingLoss(MixModule, nn.CosineEmbeddingLoss):
        __init__ = functools.partial(nn.CosineEmbeddingLoss.__init__)


with suppress(AttributeError):

    class CosineSimilarity(MixModule, nn.CosineSimilarity):
        __init__ = functools.partial(nn.CosineSimilarity.__init__)


with suppress(AttributeError):

    class CrossEntropyLoss(MixModule, nn.CrossEntropyLoss):
        __init__ = functools.partial(nn.CrossEntropyLoss.__init__)


with suppress(AttributeError):

    class CrossMapLRN2d(MixModule, nn.CrossMapLRN2d):
        __init__ = functools.partial(nn.CrossMapLRN2d.__init__)


with suppress(AttributeError):

    class DataParallel(MixModule, nn.DataParallel):
        __init__ = functools.partial(nn.DataParallel.__init__)


with suppress(AttributeError):

    class Dropout(MixModule, nn.Dropout):
        __init__ = functools.partial(nn.Dropout.__init__)


with suppress(AttributeError):

    class Dropout1d(MixModule, nn.Dropout1d):
        __init__ = functools.partial(nn.Dropout1d.__init__)


with suppress(AttributeError):

    class Dropout2d(MixModule, nn.Dropout2d):
        __init__ = functools.partial(nn.Dropout2d.__init__)


with suppress(AttributeError):

    class Dropout3d(MixModule, nn.Dropout3d):
        __init__ = functools.partial(nn.Dropout3d.__init__)


with suppress(AttributeError):

    class ELU(MixModule, nn.ELU):
        __init__ = functools.partial(nn.ELU.__init__)


with suppress(AttributeError):

    class Embedding(MixModule, nn.Embedding):
        __init__ = functools.partial(nn.Embedding.__init__)


with suppress(AttributeError):

    class EmbeddingBag(MixModule, nn.EmbeddingBag):
        __init__ = functools.partial(nn.EmbeddingBag.__init__)


with suppress(AttributeError):

    class FeatureAlphaDropout(MixModule, nn.FeatureAlphaDropout):
        __init__ = functools.partial(nn.FeatureAlphaDropout.__init__)


with suppress(AttributeError):

    class Flatten(MixModule, nn.Flatten):
        __init__ = functools.partial(nn.Flatten.__init__)


with suppress(AttributeError):

    class Fold(MixModule, nn.Fold):
        __init__ = functools.partial(nn.Fold.__init__)


with suppress(AttributeError):

    class FractionalMaxPool2d(MixModule, nn.FractionalMaxPool2d):
        __init__ = functools.partial(nn.FractionalMaxPool2d.__init__)


with suppress(AttributeError):

    class FractionalMaxPool3d(MixModule, nn.FractionalMaxPool3d):
        __init__ = functools.partial(nn.FractionalMaxPool3d.__init__)


with suppress(AttributeError):

    class GELU(MixModule, nn.GELU):
        __init__ = functools.partial(nn.GELU.__init__)


with suppress(AttributeError):

    class GLU(MixModule, nn.GLU):
        __init__ = functools.partial(nn.GLU.__init__)


with suppress(AttributeError):

    class GRU(MixModule, nn.GRU):
        __init__ = functools.partial(nn.GRU.__init__)


with suppress(AttributeError):

    class GRUCell(MixModule, nn.GRUCell):
        __init__ = functools.partial(nn.GRUCell.__init__)


with suppress(AttributeError):

    class GaussianNLLLoss(MixModule, nn.GaussianNLLLoss):
        __init__ = functools.partial(nn.GaussianNLLLoss.__init__)


with suppress(AttributeError):

    class GroupNorm(MixModule, nn.GroupNorm):
        __init__ = functools.partial(nn.GroupNorm.__init__)


with suppress(AttributeError):

    class Hardshrink(MixModule, nn.Hardshrink):
        __init__ = functools.partial(nn.Hardshrink.__init__)


with suppress(AttributeError):

    class Hardsigmoid(MixModule, nn.Hardsigmoid):
        __init__ = functools.partial(nn.Hardsigmoid.__init__)


with suppress(AttributeError):

    class Hardswish(MixModule, nn.Hardswish):
        __init__ = functools.partial(nn.Hardswish.__init__)


with suppress(AttributeError):

    class Hardtanh(MixModule, nn.Hardtanh):
        __init__ = functools.partial(nn.Hardtanh.__init__)


with suppress(AttributeError):

    class HingeEmbeddingLoss(MixModule, nn.HingeEmbeddingLoss):
        __init__ = functools.partial(nn.HingeEmbeddingLoss.__init__)


with suppress(AttributeError):

    class HuberLoss(MixModule, nn.HuberLoss):
        __init__ = functools.partial(nn.HuberLoss.__init__)


with suppress(AttributeError):

    class Identity(MixModule, nn.Identity):
        __init__ = functools.partial(nn.Identity.__init__)


with suppress(AttributeError):

    class InstanceNorm1d(MixModule, nn.InstanceNorm1d):
        __init__ = functools.partial(nn.InstanceNorm1d.__init__)


with suppress(AttributeError):

    class InstanceNorm2d(MixModule, nn.InstanceNorm2d):
        __init__ = functools.partial(nn.InstanceNorm2d.__init__)


with suppress(AttributeError):

    class InstanceNorm3d(MixModule, nn.InstanceNorm3d):
        __init__ = functools.partial(nn.InstanceNorm3d.__init__)


with suppress(AttributeError):

    class KLDivLoss(MixModule, nn.KLDivLoss):
        __init__ = functools.partial(nn.KLDivLoss.__init__)


with suppress(AttributeError):

    class L1Loss(MixModule, nn.L1Loss):
        __init__ = functools.partial(nn.L1Loss.__init__)


with suppress(AttributeError):

    class LPPool1d(MixModule, nn.LPPool1d):
        __init__ = functools.partial(nn.LPPool1d.__init__)


with suppress(AttributeError):

    class LPPool2d(MixModule, nn.LPPool2d):
        __init__ = functools.partial(nn.LPPool2d.__init__)


with suppress(AttributeError):

    class LSTM(MixModule, nn.LSTM):
        __init__ = functools.partial(nn.LSTM.__init__)


with suppress(AttributeError):

    class LSTMCell(MixModule, nn.LSTMCell):
        __init__ = functools.partial(nn.LSTMCell.__init__)


with suppress(AttributeError):

    class LayerNorm(MixModule, nn.LayerNorm):
        __init__ = functools.partial(nn.LayerNorm.__init__)


with suppress(AttributeError):

    class LazyBatchNorm1d(MixModule, nn.LazyBatchNorm1d):
        __init__ = functools.partial(nn.LazyBatchNorm1d.__init__)


with suppress(AttributeError):

    class LazyBatchNorm2d(MixModule, nn.LazyBatchNorm2d):
        __init__ = functools.partial(nn.LazyBatchNorm2d.__init__)


with suppress(AttributeError):

    class LazyBatchNorm3d(MixModule, nn.LazyBatchNorm3d):
        __init__ = functools.partial(nn.LazyBatchNorm3d.__init__)


with suppress(AttributeError):

    class LazyInstanceNorm1d(MixModule, nn.LazyInstanceNorm1d):
        __init__ = functools.partial(nn.LazyInstanceNorm1d.__init__)


with suppress(AttributeError):

    class LazyInstanceNorm2d(MixModule, nn.LazyInstanceNorm2d):
        __init__ = functools.partial(nn.LazyInstanceNorm2d.__init__)


with suppress(AttributeError):

    class LazyInstanceNorm3d(MixModule, nn.LazyInstanceNorm3d):
        __init__ = functools.partial(nn.LazyInstanceNorm3d.__init__)


with suppress(AttributeError):

    class LazyLinear(MixModule, nn.LazyLinear):
        __init__ = functools.partial(nn.LazyLinear.__init__)


with suppress(AttributeError):

    class LeakyReLU(MixModule, nn.LeakyReLU):
        __init__ = functools.partial(nn.LeakyReLU.__init__)


with suppress(AttributeError):

    class Linear(MixModule, nn.Linear):
        __init__ = functools.partial(nn.Linear.__init__)


with suppress(AttributeError):

    class LocalResponseNorm(MixModule, nn.LocalResponseNorm):
        __init__ = functools.partial(nn.LocalResponseNorm.__init__)


with suppress(AttributeError):

    class LogSigmoid(MixModule, nn.LogSigmoid):
        __init__ = functools.partial(nn.LogSigmoid.__init__)


with suppress(AttributeError):

    class LogSoftmax(MixModule, nn.LogSoftmax):
        __init__ = functools.partial(nn.LogSoftmax.__init__)


with suppress(AttributeError):

    class MSELoss(MixModule, nn.MSELoss):
        __init__ = functools.partial(nn.MSELoss.__init__)


with suppress(AttributeError):

    class MarginRankingLoss(MixModule, nn.MarginRankingLoss):
        __init__ = functools.partial(nn.MarginRankingLoss.__init__)


with suppress(AttributeError):

    class MaxPool1d(MixModule, nn.MaxPool1d):
        __init__ = functools.partial(nn.MaxPool1d.__init__)


with suppress(AttributeError):

    class MaxPool2d(MixModule, nn.MaxPool2d):
        __init__ = functools.partial(nn.MaxPool2d.__init__)


with suppress(AttributeError):

    class MaxPool3d(MixModule, nn.MaxPool3d):
        __init__ = functools.partial(nn.MaxPool3d.__init__)


with suppress(AttributeError):

    class MaxUnpool1d(MixModule, nn.MaxUnpool1d):
        __init__ = functools.partial(nn.MaxUnpool1d.__init__)


with suppress(AttributeError):

    class MaxUnpool2d(MixModule, nn.MaxUnpool2d):
        __init__ = functools.partial(nn.MaxUnpool2d.__init__)


with suppress(AttributeError):

    class MaxUnpool3d(MixModule, nn.MaxUnpool3d):
        __init__ = functools.partial(nn.MaxUnpool3d.__init__)


with suppress(AttributeError):

    class Mish(MixModule, nn.Mish):
        __init__ = functools.partial(nn.Mish.__init__)


with suppress(AttributeError):

    class ModuleDict(MixModule, nn.ModuleDict):
        __init__ = functools.partial(nn.ModuleDict.__init__)


with suppress(AttributeError):

    class ModuleList(MixModule, nn.ModuleList):
        __init__ = functools.partial(nn.ModuleList.__init__)


with suppress(AttributeError):

    class MultiLabelMarginLoss(MixModule, nn.MultiLabelMarginLoss):
        __init__ = functools.partial(nn.MultiLabelMarginLoss.__init__)


with suppress(AttributeError):

    class MultiLabelSoftMarginLoss(MixModule, nn.MultiLabelSoftMarginLoss):
        __init__ = functools.partial(nn.MultiLabelSoftMarginLoss.__init__)


with suppress(AttributeError):

    class MultiMarginLoss(MixModule, nn.MultiMarginLoss):
        __init__ = functools.partial(nn.MultiMarginLoss.__init__)


with suppress(AttributeError):

    class MultiheadAttention(MixModule, nn.MultiheadAttention):
        __init__ = functools.partial(nn.MultiheadAttention.__init__)


with suppress(AttributeError):

    class NLLLoss(MixModule, nn.NLLLoss):
        __init__ = functools.partial(nn.NLLLoss.__init__)


with suppress(AttributeError):

    class NLLLoss2d(MixModule, nn.NLLLoss2d):
        __init__ = functools.partial(nn.NLLLoss2d.__init__)


with suppress(AttributeError):

    class PReLU(MixModule, nn.PReLU):
        __init__ = functools.partial(nn.PReLU.__init__)


with suppress(AttributeError):

    class PairwiseDistance(MixModule, nn.PairwiseDistance):
        __init__ = functools.partial(nn.PairwiseDistance.__init__)


with suppress(AttributeError):

    class ParameterDict(MixModule, nn.ParameterDict):
        __init__ = functools.partial(nn.ParameterDict.__init__)


with suppress(AttributeError):

    class ParameterList(MixModule, nn.ParameterList):
        __init__ = functools.partial(nn.ParameterList.__init__)


with suppress(AttributeError):

    class PixelShuffle(MixModule, nn.PixelShuffle):
        __init__ = functools.partial(nn.PixelShuffle.__init__)


with suppress(AttributeError):

    class PixelUnshuffle(MixModule, nn.PixelUnshuffle):
        __init__ = functools.partial(nn.PixelUnshuffle.__init__)


with suppress(AttributeError):

    class PoissonNLLLoss(MixModule, nn.PoissonNLLLoss):
        __init__ = functools.partial(nn.PoissonNLLLoss.__init__)


with suppress(AttributeError):

    class RNN(MixModule, nn.RNN):
        __init__ = functools.partial(nn.RNN.__init__)


with suppress(AttributeError):

    class RNNBase(MixModule, nn.RNNBase):
        __init__ = functools.partial(nn.RNNBase.__init__)


with suppress(AttributeError):

    class RNNCell(MixModule, nn.RNNCell):
        __init__ = functools.partial(nn.RNNCell.__init__)


with suppress(AttributeError):

    class RNNCellBase(MixModule, nn.RNNCellBase):
        __init__ = functools.partial(nn.RNNCellBase.__init__)


with suppress(AttributeError):

    class RReLU(MixModule, nn.RReLU):
        __init__ = functools.partial(nn.RReLU.__init__)


with suppress(AttributeError):

    class ReLU(MixModule, nn.ReLU):
        __init__ = functools.partial(nn.ReLU.__init__)


with suppress(AttributeError):

    class ReLU6(MixModule, nn.ReLU6):
        __init__ = functools.partial(nn.ReLU6.__init__)


with suppress(AttributeError):

    class ReflectionPad1d(MixModule, nn.ReflectionPad1d):
        __init__ = functools.partial(nn.ReflectionPad1d.__init__)


with suppress(AttributeError):

    class ReflectionPad2d(MixModule, nn.ReflectionPad2d):
        __init__ = functools.partial(nn.ReflectionPad2d.__init__)


with suppress(AttributeError):

    class ReflectionPad3d(MixModule, nn.ReflectionPad3d):
        __init__ = functools.partial(nn.ReflectionPad3d.__init__)


with suppress(AttributeError):

    class ReplicationPad1d(MixModule, nn.ReplicationPad1d):
        __init__ = functools.partial(nn.ReplicationPad1d.__init__)


with suppress(AttributeError):

    class ReplicationPad2d(MixModule, nn.ReplicationPad2d):
        __init__ = functools.partial(nn.ReplicationPad2d.__init__)


with suppress(AttributeError):

    class ReplicationPad3d(MixModule, nn.ReplicationPad3d):
        __init__ = functools.partial(nn.ReplicationPad3d.__init__)


with suppress(AttributeError):

    class SELU(MixModule, nn.SELU):
        __init__ = functools.partial(nn.SELU.__init__)


with suppress(AttributeError):

    class Sequential(MixModule, nn.Sequential):
        build_mode = BuildMode.WITH_ARGS
        __init__ = functools.partial(nn.Sequential.__init__)  # type: ignore


with suppress(AttributeError):

    class SiLU(MixModule, nn.SiLU):
        __init__ = functools.partial(nn.SiLU.__init__)


with suppress(AttributeError):

    class Sigmoid(MixModule, nn.Sigmoid):
        __init__ = functools.partial(nn.Sigmoid.__init__)


with suppress(AttributeError):

    class SmoothL1Loss(MixModule, nn.SmoothL1Loss):
        __init__ = functools.partial(nn.SmoothL1Loss.__init__)


with suppress(AttributeError):

    class SoftMarginLoss(MixModule, nn.SoftMarginLoss):
        __init__ = functools.partial(nn.SoftMarginLoss.__init__)


with suppress(AttributeError):

    class Softmax(MixModule, nn.Softmax):
        __init__ = functools.partial(nn.Softmax.__init__)


with suppress(AttributeError):

    class Softmax2d(MixModule, nn.Softmax2d):
        __init__ = functools.partial(nn.Softmax2d.__init__)


with suppress(AttributeError):

    class Softmin(MixModule, nn.Softmin):
        __init__ = functools.partial(nn.Softmin.__init__)


with suppress(AttributeError):

    class Softplus(MixModule, nn.Softplus):
        __init__ = functools.partial(nn.Softplus.__init__)


with suppress(AttributeError):

    class Softshrink(MixModule, nn.Softshrink):
        __init__ = functools.partial(nn.Softshrink.__init__)


with suppress(AttributeError):

    class Softsign(MixModule, nn.Softsign):
        __init__ = functools.partial(nn.Softsign.__init__)


with suppress(AttributeError):

    class SyncBatchNorm(MixModule, nn.SyncBatchNorm):
        __init__ = functools.partial(nn.SyncBatchNorm.__init__)


with suppress(AttributeError):

    class Tanh(MixModule, nn.Tanh):
        __init__ = functools.partial(nn.Tanh.__init__)


with suppress(AttributeError):

    class Tanhshrink(MixModule, nn.Tanhshrink):
        __init__ = functools.partial(nn.Tanhshrink.__init__)


with suppress(AttributeError):

    class Threshold(MixModule, nn.Threshold):
        __init__ = functools.partial(nn.Threshold.__init__)


with suppress(AttributeError):

    class Transformer(MixModule, nn.Transformer):
        __init__ = functools.partial(nn.Transformer.__init__)


with suppress(AttributeError):

    class TransformerDecoder(MixModule, nn.TransformerDecoder):
        __init__ = functools.partial(nn.TransformerDecoder.__init__)


with suppress(AttributeError):

    class TransformerDecoderLayer(MixModule, nn.TransformerDecoderLayer):
        __init__ = functools.partial(nn.TransformerDecoderLayer.__init__)


with suppress(AttributeError):

    class TransformerEncoder(MixModule, nn.TransformerEncoder):
        __init__ = functools.partial(nn.TransformerEncoder.__init__)


with suppress(AttributeError):

    class TransformerEncoderLayer(MixModule, nn.TransformerEncoderLayer):
        __init__ = functools.partial(nn.TransformerEncoderLayer.__init__)


with suppress(AttributeError):

    class TripletMarginLoss(MixModule, nn.TripletMarginLoss):
        __init__ = functools.partial(nn.TripletMarginLoss.__init__)


with suppress(AttributeError):

    class TripletMarginWithDistanceLoss(MixModule, nn.TripletMarginWithDistanceLoss):
        __init__ = functools.partial(nn.TripletMarginWithDistanceLoss.__init__)


with suppress(AttributeError):

    class Unflatten(MixModule, nn.Unflatten):
        __init__ = functools.partial(nn.Unflatten.__init__)


with suppress(AttributeError):

    class Unfold(MixModule, nn.Unfold):
        __init__ = functools.partial(nn.Unfold.__init__)


with suppress(AttributeError):

    class Upsample(MixModule, nn.Upsample):
        __init__ = functools.partial(nn.Upsample.__init__)


with suppress(AttributeError):

    class UpsamplingBilinear2d(MixModule, nn.UpsamplingBilinear2d):
        __init__ = functools.partial(nn.UpsamplingBilinear2d.__init__)


with suppress(AttributeError):

    class UpsamplingNearest2d(MixModule, nn.UpsamplingNearest2d):
        __init__ = functools.partial(nn.UpsamplingNearest2d.__init__)


with suppress(AttributeError):

    class ZeroPad2d(MixModule, nn.ZeroPad2d):
        __init__ = functools.partial(nn.ZeroPad2d.__init__)



if __name__ != "__main__":
    raise ImportError("This file is not intended to be imported!")

import inspect
import subprocess

from torch import nn

DOCS = """
This file is automatically generated by torchmix and should not be changed.

This module contains wrapped versions of the objects in the PyTorch `nn`
module, intended to provide a more user-friendly interface and added
configuration management functionality.

These classes can be used as a drop-in replacement for the original PyTorch
`nn` module, allowing you to access the full configuration of the module
through the config attribute.
"""


def lazy_conv_template(name):
    f"""with suppress(AttributeError):
    class {name}(MixModule, nn.{name}):
        __init__ = functools.partial(nn.{name}.__init__)  # type: ignore[misc]"""


def sequential_template(name):
    return f"""with suppress(AttributeError):
    class {name}(MixModule, nn.{name}):
        build_mode = BuildMode.WITH_ARGS
        __init__ = functools.partial(nn.{name}.__init__)  # type: ignore"""


def module_template(name):
    return f"""with suppress(AttributeError):
    class {name}(MixModule, nn.{name}):
        __init__ = functools.partial(nn.{name}.__init__)"""


with open("torchmix/nn/__init__.py", "w") as f:
    f.write('"""')
    f.write(DOCS)
    f.write('"""\n\n')

    def write(line=None):
        return f.write(line + "\n" if line else "\n")

    write("from contextlib import suppress")
    write("import functools")
    write()
    write("from torch import nn")
    write()
    write("from torchmix.core._builds import BuildMode")
    write("from torchmix.core._module import MixModule")
    write()
    write()
    write("Module = nn.Module")
    write("Parameter = nn.Parameter")
    write("UninitializedBuffer = nn.UninitializedBuffer")
    write("UninitializedParameter = nn.UninitializedParameter")
    write()

    for name, obj in inspect.getmembers(nn):
        if name in ("Module", "Parameter") or name.startswith(("_", "__")):
            continue
        try:
            if issubclass(obj, nn.Module):
                pass
        except:  # noqa: E722
            if name not in (
                "common_types",
                "grad",
                "intrinsic",
                "qat",
                "quantizable",
                "quantized",
            ):
                write("with suppress(AttributeError):")
                write(f"    {name} = nn.{name}\n")

    processed = set()
    for name, obj in inspect.getmembers(nn):
        if name in ("Module", "Parameter") or name.startswith(("_", "__")):
            continue

        try:
            template: str
            if name == "Sequential":
                template = sequential_template(name)
            elif name.startswith("LazyConv"):
                template = lazy_conv_template(name)
            elif issubclass(obj, nn.Module):
                template = module_template(name)
            else:
                raise ValueError(f"{name} not handled")

            write()
            write()
            write(template)

            processed.add(name)

        except:  # noqa: E722
            pass

try:
    subprocess.run(["poetry", "run", "black", "-l", "78", "torchmix/nn/__init__.py"])
except:  # noqa: E722
    pass



import contextlib
import unittest.mock


@contextlib.contextmanager
def no_parameters():
    """
    Context manager that temporarily patches `MixModul.__new__` method to
    skip calling the original `__init__` method of `MixModul` instances.

    This can be useful if you want to create a `MixModul` instance
    solely for the purpose of generating its configuration, without actually
    instantiating the module.

    Example:
        >>> with torchmix.no_parameters():
        ...     model = torchmix.nn.Linear(100, 200)
        >>> print(model._parameters)
        AttributeError: 'Linear' object has no attribute '_parameters'
    """
    with unittest.mock.patch("torchmix.core._module.NO_PARAMS", True):
        yield



from .containers import Add, Attach, PostNorm, PreNorm, Repeat
from .layers import (
    AvgPool,
    ChannelMixer,
    Extract,
    PatchEmbed,
    PositionEmbed,
    Token,
    TokenMixer,
)
from .layers.attentions import SelfAttention, WindowAttention

__all__ = [
    "Attach",
    "Add",
    "Repeat",
    "PreNorm",
    "PostNorm",
    "AvgPool",
    "PatchEmbed",
    "PositionEmbed",
    "ChannelMixer",
    "TokenMixer",
    "SelfAttention",
    "WindowAttention",
    "Token",
    "Extract",
]



from enum import Enum, auto

from hydra_zen import make_custom_builds_fn


class BuildMode(Enum):
    WITH_ARGS = auto()
    WITHOUT_ARGS = auto()


builds = make_custom_builds_fn(
    populate_full_signature=True,
)



from einops.layers._einmix import _EinmixMixin
from einops.layers.torch import EinMix as _EinMix
from einops.layers.torch import Rearrange as _Rearrange
from einops.layers.torch import RearrangeMixin as _RearrangeMixin
from einops.layers.torch import Reduce as _Reduce
from einops.layers.torch import ReduceMixin as _ReduceMixin

from torchmix.core._module import MixModule


class EinMix(MixModule, _EinMix):
    __init__ = _EinmixMixin.__init__  # type: ignore


class Rearrange(MixModule, _Rearrange):
    __init__ = _RearrangeMixin.__init__  # type: ignore


class Reduce(MixModule, _Reduce):
    __init__ = _ReduceMixin.__init__  # type: ignore



import math

import torch
from hydra_zen.typing import Partial
from jaxtyping import Float
from torch import Tensor

from torchmix import nn
from torchmix.core._module import MixModule
from torchmix.third_party.einops import EinMix, Rearrange, Reduce

__all__ = [
    "AvgPool",
    "Extract",
    "PatchEmbed",
    "PositionEmbed",
    "ChannelMixer",
    "TokenMixer",
    "Token",
]


class AvgPool(MixModule):
    def __init__(self):
        self.pool = Reduce("b n d -> b d", reduction="mean")

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b d"]:
        return self.pool(x)


class PatchEmbed(MixModule):
    def __init__(
        self,
        patch_size: int = 16,
        channels: int = 3,
        dim: int = 768,
    ):
        self.projection = EinMix(
            "b c (h ph) (w pw) -> b (h w) d",
            weight_shape="c ph pw d",
            bias_shape="d",
            ph=patch_size,
            pw=patch_size,
            c=channels,
            d=dim,
        )

    def forward(self, x: Float[Tensor, "b c h w"]) -> Float[Tensor, "b n d"]:
        return self.projection(x)


class ChannelMixer(MixModule):
    def __init__(
        self,
        act_layer: Partial[MixModule],
        dim: int = 1024,
        expansion_factor: float = 4,
    ):
        self.block = nn.Sequential(
            EinMix(
                "b n d_in -> b n d_out",
                weight_shape="d_in d_out",
                bias_shape="d_out",
                d_in=dim,
                d_out=int(dim * expansion_factor),
            ),
            act_layer(),
            EinMix(
                "b n d_out -> b n d_in",
                weight_shape="d_out d_in",
                bias_shape="d_in",
                d_in=dim,
                d_out=int(dim * expansion_factor),
            ),
        )

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b n d"]:
        return self.block(x)


class TokenMixer(MixModule):
    def __init__(
        self,
        act_layer: Partial[MixModule],
        seq_length: int = 196,
        expansion_factor: float = 0.5,
    ):
        self.block = nn.Sequential(
            EinMix(
                "b n_in d -> b n_out d",
                weight_shape="n_in n_out",
                bias_shape="n_out",
                n_in=seq_length,
                n_out=int(seq_length * expansion_factor),
            ),
            act_layer(),
            EinMix(
                "b n_out d -> b n_in d",
                weight_shape="n_out n_in",
                bias_shape="n_in",
                n_in=seq_length,
                n_out=int(seq_length * expansion_factor),
            ),
        )

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b n d"]:
        return self.block(x)


class Token(MixModule):
    def __init__(
        self,
        dim: int = 1024,
    ):
        self.class_token = nn.Parameter(torch.zeros(dim))

    def forward(self, *_) -> Float[Tensor, " d"]:
        return self.class_token


class PositionEmbed(MixModule):
    def __init__(
        self,
        seq_length: int = 197,
        dim: int = 1024,
    ):
        self.pos_embed = nn.Parameter(torch.randn(seq_length, dim)) * 0.02

    def forward(self, *_) -> Float[Tensor, "n d"]:
        return self.pos_embed


class Extract(MixModule):
    def __init__(self, index: int):
        self.index = index

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b d"]:
        return x[:, self.index, :]


class PatchMerging(MixModule):
    def __init__(self, dim: int = 96):
        self.proj = nn.Sequential(
            Rearrange(
                "b (h ph) (w pw) d_in -> b (h w) (ph pw d_in)",
                ph=2,
                pw=2,
            ),
            nn.LayerNorm(dim * 4),
            EinMix(
                "b n d_in -> b n d_out",
                weight_shape="d_in d_out",
                bias_shape="d_out",
                d_in=dim,
                d_out=dim * 2,
            ),
        )

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b n/4 d*2"]:
        B, N, D = x.shape
        H = W = round(math.sqrt(N))
        x = x.view(B, H, W, D)
        return self.proj(x)



from typing import Iterator

from einops import pack, repeat
from torch import Tensor

from torchmix import nn
from torchmix.core._builds import BuildMode
from torchmix.core._module import MixModule

__all__ = ["Attach", "Add", "Repeat", "PreNorm", "PostNorm"]


class Attach(MixModule):
    build_mode = BuildMode.WITH_ARGS

    def __init__(self, *blocks: MixModule):
        for idx, block in enumerate(blocks):
            self.add_module(str(idx), block)

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[MixModule]:
        return iter(self._modules.values())  # type: ignore

    def forward(self, x: Tensor) -> Tensor:
        tokens = []
        for idx, block in enumerate(self):
            _token: Tensor = block(x)
            if idx == 0:
                if _token.dim() != 3:
                    raise ValueError(
                        "Invalid tensor dimensions: "
                        f"expected 3 dimensions, got {_token.dim()}"
                    )
                batch_size, *_ = _token.shape

            if _token.dim() == 1:
                tokens.append(repeat(_token, "d -> b d", b=batch_size))
            elif _token.dim() == 2:
                tokens.append(repeat(_token, "n d -> b n d", b=batch_size))
            elif _token.dim() == 3:
                tokens.append(_token)
            else:
                raise ValueError(
                    "Invalid tensor dimensions: "
                    f"expected at most 3 dimensions, got {_token.dim()}"
                )

        x, _packed_shape = pack(tokens, "b * c")

        return x


class Add(MixModule):
    build_mode = BuildMode.WITH_ARGS

    def __init__(self, *blocks: MixModule):
        for idx, block in enumerate(blocks):
            self.add_module(str(idx), block)

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[MixModule]:
        return iter(self._modules.values())  # type: ignore

    def forward(self, x: Tensor) -> Tensor:
        _x: Tensor
        for idx, module in enumerate(self):
            if idx == 0:
                _x = module(x)
            else:
                _x.add_(module(x))
        return _x


class Repeat(MixModule):
    def __init__(
        self,
        block: MixModule,
        depth: int = 8,
    ) -> None:
        for idx in range(depth):
            self.add_module(
                str(idx),
                block.instantiate(),
            )

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[MixModule]:
        return iter(self._modules.values())  # type: ignore

    def forward(self, x: Tensor) -> Tensor:
        for module in self:
            x = module(x)
        return x


class PreNorm(MixModule):
    def __init__(self, block: MixModule, dim: int = 1024):
        self.block = block
        self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor) -> Tensor:
        return x + self.block(self.norm(x))


class PostNorm(MixModule):
    def __init__(self, block: MixModule, dim: int = 1024):
        self.block = block
        self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor) -> Tensor:
        return self.norm(x + self.block(x))



import math

from einops import einsum, rearrange, unpack
from jaxtyping import Float
from torch import Tensor

from torchmix.core._module import MixModule
from torchmix.third_party.einops import EinMix

__all__ = [
    "SelfAttention",
    "WindowAttention",
]


class SelfAttention(MixModule):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        head_dim: int,
    ):
        self.inner_dim = num_heads * head_dim
        self.scale: float = head_dim**-0.5
        self.num_heads: float = num_heads

        self._to_qkv = EinMix(
            "b n d_in -> b n d_out",
            weight_shape="d_in d_out",
            bias_shape="d_out",
            d_in=dim,
            d_out=3 * self.inner_dim,
        )

        self._proj = EinMix(
            "b n d_out -> b n d_in",
            weight_shape="d_in d_out",
            bias_shape="d_in",
            d_in=dim,
            d_out=self.inner_dim,
        )

    def to_qkv(
        self, x: Float[Tensor, "b n d"]
    ) -> tuple[
        Float[Tensor, "b n d"], Float[Tensor, "b n d"], Float[Tensor, "b n d"]
    ]:
        query, key, value = unpack(
            self._to_qkv(x),
            [[self.inner_dim], [self.inner_dim], [self.inner_dim]],
            "b n *",
        )
        return query, key, value

    def split_qkv(
        self,
        query: Float[Tensor, "b n d_in"],
        key: Float[Tensor, "b n d_in"],
        value: Float[Tensor, "b n d_in"],
    ) -> tuple[
        Float[Tensor, "b h n d_out"],
        Float[Tensor, "b h n d_out"],
        Float[Tensor, "b h n d_out"],
    ]:
        query, key, value = map(
            lambda x: rearrange(
                x,
                "b n (h d) -> b h n d",
                h=self.num_heads,
            ),
            (query, key, value),
        )
        return query, key, value

    def attention(
        self,
        query: Float[Tensor, "... n d"],
        key: Float[Tensor, "... n d"],
        value: Float[Tensor, "... n d"],
    ) -> Float[Tensor, "... n d"]:
        dots = einsum(query, key, "... q d, ... k d -> ... q k") * self.scale
        attention = dots.softmax(dim=-1)
        out = einsum(attention, value, "... q k, ... k d -> ... q d")
        return out

    def collect_heads(
        self, out: Float[Tensor, "b h n d"]
    ) -> Float[Tensor, "b n h*d"]:
        return rearrange(out, "b h n d -> b n (h d)")

    def forward(self, x: Float[Tensor, "b n d"]) -> Float[Tensor, "b n d"]:
        q, k, v = self.to_qkv(x)
        q, k, v = self.split_qkv(q, k, v)
        out = self.attention(q, k, v)
        out = self.collect_heads(out)
        out = self._proj(out)
        return out


class WindowAttention(SelfAttention):
    def __init__(
        self,
        dim: int = 96,
        window_size: int = 8,
        num_heads: int = 8,
        head_dim: int = 64,
    ):
        super().__init__(
            dim=dim,
            num_heads=num_heads,
            head_dim=head_dim,
        )
        self.window_size: int = window_size

    def split_qkv(
        self,
        query: Float[Tensor, "b n d"],
        key: Float[Tensor, "b n d"],
        value: Float[Tensor, "b n d"],
    ) -> tuple[
        Float[Tensor, "b n d"],
        Float[Tensor, "b n d"],
        Float[Tensor, "b n d"],
    ]:
        _batch_size, _seq_length, _inner_dim = query.shape
        patch_size = round(math.sqrt(_seq_length))

        query, key, value = map(
            lambda x: rearrange(
                x,
                "b (h ph w pw) (hd d) -> b h w hd (ph pw) d",
                ph=self.window_size,
                pw=self.window_size,
                h=patch_size // self.window_size,
                w=patch_size // self.window_size,
                hd=self.num_heads,
            ),
            (query, key, value),
        )
        return query, key, value

    def collect_heads(
        self, out: Float[Tensor, "b h n d_out"]
    ) -> Float[Tensor, "b n d_in"]:
        return rearrange(
            out,
            "b h w hd (ph pw) d -> b (h ph w pw) (hd d)",
            ph=self.window_size,
            pw=self.window_size,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchmix',
 'torchmix.components',
 'torchmix.components.layers',
 'torchmix.core',
 'torchmix.nn',
 'torchmix.third_party']

package_data = \
{'': ['*']}

install_requires = \
['einops>=0.6.0,<0.7.0',
 'hydra-core>=1.3.1,<2.0.0',
 'hydra-zen>=0.8.0,<0.9.0',
 'jaxtyping>=0.2.11,<0.3.0']

setup_kwargs = {
    'name': 'torchmix',
    'version': '0.1.0rc2',
    'description': '',
    'long_description': '<h1 align="center">torchmix</h1>\n\n<h3 align="center">The missing component library for PyTorch</h3>\n\n<br />\n\nWelcome to torchmix, a collection of PyTorch modules that aims to reduce boilerplate and improve code modularity.\n\n**Please note: `torchmix` is currently in development and has not been tested for production use. The API may change at any time.**\n\n<br />\n\n## Usage\n\nTo use `torchmix`, simply import the desired components:\n\n```python\nimport torchmix.nn as nn  # Wrapped version of torch.nn\nfrom torchmix import (\n    Add,\n    Attach,\n    AvgPool,\n    ChannelMixer,\n    Extract,\n    PatchEmbed,\n    PositionEmbed,\n    PreNorm,\n    Repeat,\n    SelfAttention,\n    Token,\n)\n```\n\nYou can simply compose this components to build more complex architecture, as shown in the following example:\n\n```python\nvit_cls = nn.Sequential(\n    Add(\n        Attach(\n            PatchEmbed(dim=1024),\n            Token(dim=1024),\n        ),\n        PositionEmbed(\n            seq_length=196 + 1,\n            dim=1024,\n        ),\n    ),\n    Repeat(\n        nn.Sequential(\n            PreNorm(\n                ChannelMixer(\n                    dim=1024,\n                    expansion_factor=4,\n                    act_layer=nn.GELU.partial(),\n                ),\n                dim=1024,\n            ),\n            PreNorm(\n                SelfAttention(\n                    dim=1024,\n                    num_heads=8,\n                    head_dim=64,\n                ),\n                dim=1024,\n            ),\n        )\n    ),\n    Extract(0),\n)\n\nvit_avg = nn.Sequential(\n    Add(\n        PatchEmbed(dim=1024),\n        PositionEmbed(\n            seq_length=196,\n            dim=1024,\n        ),\n    ),\n    Repeat(\n        nn.Sequential(\n            PreNorm(\n                ChannelMixer(\n                    dim=1024,\n                    expansion_factor=4,\n                    act_layer=nn.GELU.partial(),\n                ),\n                dim=1024,\n            ),\n            PreNorm(\n                SelfAttention(\n                    dim=1024,\n                    num_heads=8,\n                    head_dim=64,\n                ),\n                dim=1024,\n            ),\n        )\n    ),\n    AvgPool(),\n)\n```\n\n### Integration with Hydra\n\nReproducibility is important, so it is always a good idea to store the configurations of your models. However, manually writing the configurations for complex, deeply nested PyTorch modules can be tedious and result in code that is difficult to understand and maintain. This is because the parent class may need to accept and pass along the parameters of its children classes, leading to a large number of arguments and strong coupling between the parent and child classes.\n\n`torchmix` simplifies this process by **auto-magically** generating the full configuration of a PyTorch module **simply by instantiating it.** This enables effortless integration with the `hydra` ecosystem, which allows for easy storage and management of module configurations.\n\nTo generate a configuration for a typical MLP using `torchmix`, for example, you can do the following:\n\n```python\nfrom torchmix import nn\n\nmodel = nn.Sequential(\n    nn.Linear(1024, 4096),\n    nn.Dropout(0.1),\n    nn.GELU(),\n    nn.Linear(4096, 1024),\n    nn.Dropout(0.1),\n)\n```\n\nYou can then store the configuration in the `hydra`\'s `ConfigStore` using:\n\n```python\nmodel.store(group="model", name="mlp")\n```\n\nAlternatively, you can export it to a YAML file if you want:\n\n```python\nmodel.export("mlp.yaml")\n```\n\nThis will generate the following configuration:\n\n```yaml\n_target_: torchmix.nn.Sequential\n_args_:\n  - _target_: torchmix.nn.Linear\n    in_features: 1024\n    out_features: 4096\n    bias: true\n    device: null\n    dtype: null\n  - _target_: torchmix.nn.Dropout\n    p: 0.1\n    inplace: false\n  - _target_: torchmix.nn.GELU\n    approximate: none\n  - _target_: torchmix.nn.Linear\n    in_features: 4096\n    out_features: 1024\n    bias: true\n    device: null\n    dtype: null\n  - _target_: torchmix.nn.Dropout\n    p: 0.1\n    inplace: false\n```\n\nYou can always instantiate the actual PyTorch module from its configuration using `hydra`\'s `instantiate` function.\n\nTo create custom modules with this functionality, simply subclass `MixModule` and define your module as you normally would:\n\n```python\nfrom torchmix import MixModule\n\nclass CustomModule(MixModule):\n    def __init__(self, num_heads, dim, depth):\n        pass\n\ncustom_module = CustomModule(16, 768, 12)\ncustom_module.store(group="model", name="custom")\n```\n\n## Documentation\n\nDocumentation is currently in progress. Please stay tuned! 🚀\n\n## Contributing\n\nWe welcome contributions to the `torchmix` library. If you have ideas for new components or suggestions for improving the library, don\'t hesitate to open an issue or start a discussion. Please note that `torchmix` is still in the prototype phase, so any contributions should be considered experimental.\n\n## License\n\n`torchmix` is licensed under the MIT License.\n',
    'author': 'junhsss',
    'author_email': 'junhsssr@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
